#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/layers/gpp_layer.hpp"
#include "caffe/util/math_functions.hpp"

namespace caffe {

using std::min;
using std::max;

template <typename Dtype>
void GPPLayer<Dtype>::LayerSetUp(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {
  PoolingParameter pool_param = this->layer_param_.pooling_param();
  neighbors_ = pool_param.neighbors();
  CHECK_GE(neighbors_, 4) << "Number of neighbors is at least 4.";
  sigma_ = pool_param.sigma();
  CHECK_GE(sigma_, 0) << "Smoothing parameter cannot be negative.";
  CHECK_LE(sigma_, 1) << "Smoothing parameter cannot be larger than 1.";
}

template <typename Dtype>
void GPPLayer<Dtype>::Reshape(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {
  CHECK_EQ(4, bottom[0]->num_axes()) << "Input must have 4 axes, "
      << "corresponding to (num, channels, height, width)";
  channels_ = bottom[0]->channels();
  height_ = bottom[0]->height();
  width_ = bottom[0]->width();
  top[0]->Reshape(bottom[0]->num(), channels_, height_, width_);
  max_idx_.Reshape(bottom[0]->num(), channels_, height_, width_);
}

template <typename Dtype>
void GPPLayer<Dtype>::Forward_cpu(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {
  const Dtype* bottom_data = bottom[0]->cpu_data();
  Dtype* top_data = top[0]->mutable_cpu_data();
  const int top_count = top[0]->count();
  int* mask = NULL;  // suppress warnings about uninitalized variables
  // Different pooling methods. We explicitly do the switch outside the for
  // loop to save time, although this results in more code.
  switch (this->layer_param_.pooling_param().pool()) {
  case PoolingParameter_PoolMethod_MAX:
    // Initialize
    mask = max_idx_.mutable_cpu_data();
    caffe_set(top_count, -1, mask);
    // The main loop
    for (int n = 0; n < bottom[0]->num(); ++n) {
      for (int c = 0; c < channels_; ++c) {
        for (int h = 0; h < height_; ++h) {
          for (int w = 0; w < width_; ++w) {
            const int index = h * width_ + w;
            top_data[index] = bottom_data[index];
            Dtype maxValue = -FLT_MAX;
            if (neighbors_ >= 4) {
              if (h > 0) {
                Dtype temp = bottom_data[index - width_] * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 0;
                }
              }
              if (w > 0) {
                Dtype temp = bottom_data[index - 1] * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 1;
                }
              }
              if (h < height_ - 1) {
                Dtype temp = bottom_data[index + width_] * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 2;
                }
              }
              if (w < width_ - 1) {
                Dtype temp = bottom_data[index + 1] * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 3;
                }
              }
            }
            if (neighbors_ >= 8) {
              if (h > 0 && w > 0) {
                Dtype temp = bottom_data[index - width_ - 1] * sigma_ * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 4;
                }
              }
              if (h > 0 && w < width_ - 1) {
                Dtype temp = bottom_data[index - width_ + 1] * sigma_ * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 5;
                }
              }
              if (h < height_ - 1 && w > 0) {
                Dtype temp = bottom_data[index + width_ - 1] * sigma_ * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 6;
                }
              }
              if (h < height_ - 1 && w < width_ - 1) {
                Dtype temp = bottom_data[index + width_ + 1] * sigma_ * sigma_;
                if (maxValue < temp) {
                  maxValue = temp;
                  mask[index] = 7;
                }
              }
            }
            top_data[index] += maxValue;
          }
        }
        // compute offset
        bottom_data += bottom[0]->offset(0, 1);
        top_data += top[0]->offset(0, 1);
        mask += top[0]->offset(0, 1);
      }
    }
    break;
  default:
    LOG(FATAL) << "Unknown pooling method.";
  }
}

template <typename Dtype>
void GPPLayer<Dtype>::Backward_cpu(const vector<Blob<Dtype>*>& top,
      const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) {
  if (!propagate_down[0]) {
    return;
  }
  const Dtype* top_diff = top[0]->cpu_diff();
  Dtype* bottom_diff = bottom[0]->mutable_cpu_diff();
  // Different pooling methods. We explicitly do the switch outside the for
  // loop to save time, although this results in more codes.
  caffe_set(bottom[0]->count(), Dtype(0), bottom_diff);
  const int* mask = NULL;  // suppress warnings about uninitialized variables
  switch (this->layer_param_.pooling_param().pool()) {
  case PoolingParameter_PoolMethod_MAX:
    mask = max_idx_.cpu_data();
    for (int n = 0; n < top[0]->num(); ++n) {
      for (int c = 0; c < channels_; ++c) {
        for (int h = 0; h < height_; ++h) {
          for (int w = 0; w < width_; ++w) {
            const int index = h * width_ + w;
            bottom_diff[index] += top_diff[index];
            if (mask[index] >= 0) {
              switch (mask[index]) {
              case 0:
                bottom_diff[index - width_] += top_diff[index] * sigma_;
                break;
              case 1:
                bottom_diff[index - 1] += top_diff[index] * sigma_;
                break;
              case 2:
                bottom_diff[index + width_] += top_diff[index] * sigma_;
                break;
              case 3:
                bottom_diff[index + 1] += top_diff[index] * sigma_;
                break;
              case 4:
                bottom_diff[index - width_ - 1] += top_diff[index] * sigma_ * sigma_;
                break;
              case 5:
                bottom_diff[index - width_ + 1] += top_diff[index] * sigma_ * sigma_;
                break;
              case 6:
                bottom_diff[index + width_ - 1] += top_diff[index] * sigma_ * sigma_;
                break;
              case 7:
                bottom_diff[index + width_ + 1] += top_diff[index] * sigma_ * sigma_;
                break;
              default:
                break;
              }
            }
          }
        }
        bottom_diff += bottom[0]->offset(0, 1);
        top_diff += top[0]->offset(0, 1);
        mask += top[0]->offset(0, 1);
      }
    }
    break;
  default:
    LOG(FATAL) << "Unknown pooling method.";
  }
}


#ifdef CPU_ONLY
STUB_GPU(GPPLayer);
#endif

INSTANTIATE_CLASS(GPPLayer);

}  // namespace caffe
