#include <algorithm>
#include <cfloat>
#include <vector>

#include "caffe/layers/gpp_layer.hpp"
#include "caffe/util/math_functions.hpp"

namespace caffe {

template <typename Dtype>
__global__ void MaxPoolForward(const int nthreads,
    const Dtype* bottom_data, Dtype* top_data, int* mask,
    const int height, const int width, int neighbors_, Dtype sigma_) {
  CUDA_KERNEL_LOOP(index, nthreads) {
    const int w = index % width;
    const int h = (index / width) % height;
    Dtype maxValue = -FLT_MAX;
    mask[index] = -1;
    if (neighbors_ >= 4) {
      if (h > 0) {
        Dtype temp = bottom_data[index - width] * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 0;
        }
      }
      if (w > 0) {
        Dtype temp = bottom_data[index - 1] * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 1;
        }
      }
      if (h < height - 1) {
        Dtype temp = bottom_data[index + width] * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 2;
        }
      }
      if (w < width - 1) {
        Dtype temp = bottom_data[index + 1] * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 3;
        }
      }
    }
    if (neighbors_ >= 8) {
      if (h > 0 && w > 0) {
        Dtype temp = bottom_data[index - width - 1] * sigma_ * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 4;
        }
      }
      if (h > 0 && w < width - 1) {
        Dtype temp = bottom_data[index - width + 1] * sigma_ * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 5;
        }
      }
      if (h < height - 1 && w > 0) {
        Dtype temp = bottom_data[index + width - 1] * sigma_ * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 6;
        }
      }
      if (h < height - 1 && w < width - 1) {
        Dtype temp = bottom_data[index + width + 1] * sigma_ * sigma_;
        if (maxValue < temp) {
          maxValue = temp;
          mask[index] = 7;
        }
      }
    }
    top_data[index] = bottom_data[index] + maxValue;
  }
}

template <typename Dtype>
void GPPLayer<Dtype>::Forward_gpu(const vector<Blob<Dtype>*>& bottom,
      const vector<Blob<Dtype>*>& top) {
  const Dtype* bottom_data = bottom[0]->gpu_data();
  Dtype* top_data = top[0]->mutable_gpu_data();
  int count = top[0]->count();
  // We'll output the mask to top[1] if it's of size >1.
  const bool use_top_mask = top.size() > 1;
  int* mask = NULL;
  switch (this->layer_param_.pooling_param().pool()) {
  case PoolingParameter_PoolMethod_MAX:
    mask = max_idx_.mutable_gpu_data();
    // NOLINT_NEXT_LINE(whitespace/operators)
    MaxPoolForward<Dtype><<<CAFFE_GET_BLOCKS(count), CAFFE_CUDA_NUM_THREADS>>>(
        count, bottom_data, top_data, mask, height_, width_, neighbors_, sigma_);
    break;
  default:
    LOG(FATAL) << "Unknown pooling method.";
  }
  CUDA_POST_KERNEL_CHECK;
}


template <typename Dtype>
__global__ void MaxPoolBackward(const int nthreads,
    const Dtype* top_diff, const int* mask, Dtype* bottom_diff,
    const int height, const int width, int neighbors_, Dtype sigma_) {
  CUDA_KERNEL_LOOP(index, nthreads) {
    const int w = index % width;
    const int h = (index / width) % height;
    bottom_diff[index] = top_diff[index];
    if (h < height - 1 && mask[index + width] == 0) {
      bottom_diff[index] += top_diff[index + width] * sigma_;
    }
    if (w < width - 1 && mask[index + 1] == 1) {
      bottom_diff[index] += top_diff[index + 1] * sigma_;
    }
    if (h > 0 && mask[index - width] == 2) {
      bottom_diff[index] += top_diff[index - width] * sigma_;
    }
    if (w > 0 && mask[index - 1] == 3) {
      bottom_diff[index] += top_diff[index - 1] * sigma_;
    }
    if (h < height - 1 && w < width - 1 && mask[index + width + 1] == 4) {
      bottom_diff[index] += top_diff[index + width + 1] * sigma_ * sigma_;
    }
    if (h < height - 1 && w > 0 && mask[index + width - 1] == 5) {
      bottom_diff[index] += top_diff[index + width - 1] * sigma_ * sigma_;
    }
    if (h > 0 && w < width - 1 && mask[index - width + 1] == 6) {
      bottom_diff[index] += top_diff[index - width + 1] * sigma_ * sigma_;
    }
    if (h > 0 && w > 0 && mask[index - width - 1] == 7) {
      bottom_diff[index] += top_diff[index - width - 1] * sigma_ * sigma_;
    }
  }
}

template <typename Dtype>
void GPPLayer<Dtype>::Backward_gpu(const vector<Blob<Dtype>*>& top,
      const vector<bool>& propagate_down, const vector<Blob<Dtype>*>& bottom) {
  if (!propagate_down[0]) {
    return;
  }
  const Dtype* top_diff = top[0]->gpu_diff();
  Dtype* bottom_diff = bottom[0]->mutable_gpu_diff();
  const int count = bottom[0]->count();
  const int* mask = NULL;
  switch (this->layer_param_.pooling_param().pool()) {
  case PoolingParameter_PoolMethod_MAX:
    mask = max_idx_.gpu_data();
    // NOLINT_NEXT_LINE(whitespace/operators)
    MaxPoolBackward<Dtype><<<CAFFE_GET_BLOCKS(count), CAFFE_CUDA_NUM_THREADS>>>(
        count, top_diff, mask, bottom_diff, height_, width_, neighbors_, sigma_);
    break;
  default:
    LOG(FATAL) << "Unknown pooling method.";
  }
  CUDA_POST_KERNEL_CHECK;
}


INSTANTIATE_LAYER_GPU_FUNCS(GPPLayer);


}  // namespace caffe
